from typing import List


class LoggedMetrics:

    logged_metrics = [
        ("loss", ["train", "val", "test"], ["all"], ["simple"]),
        ("acc", ["train", "val", "test"], ["all"], ["simple"]),
        (
            "precision",
            ["val", "test"],
            ["all", "fil"],
            ["micro", "macro"],
        ),
        ("precision", ["val", "test"], ["chk", "ind"], ["micro"]),
        ("recall", ["val", "test"], ["all", "fil"], ["micro", "macro"]),
        ("recall", ["val", "test"], ["chk", "ind"], ["micro"]),
        ("f1", ["val", "test"], ["all", "fil"], ["micro", "macro"]),
        ("f1", ["val", "test"], ["chk", "ind"], ["micro"]),
    ]

    def __init__(self):
        for metric, phase_list, tags_list, micro_macro_list in self.logged_metrics:
            assert isinstance(metric, str)
            assert isinstance(phase_list, list)
            assert isinstance(tags_list, list)
            assert isinstance(micro_macro_list, list)

    def get_metrics(
        self,
        tag_group: List[str] = None,
        phase_group: List[str] = None,
        micro_macro_group: List[str] = None,
        exclude: List[str] = None,
    ) -> List[str]:
        """
        get metrics, filtered
        ---------------------
        :param tag_group:          [list] of tags that are required, e.g. ['all']
        :param phase_group:        [list] of phases that are required, e.g. ['train']
        :param micro_macro_group:  [list] simple/micro/macro that is required, e.g. ['micro', 'macro']
        :param exclude:            [list] of metrics to exclude, e.g. ['loss']
        :return: filtered_metrics: [list] of [str], e.g. ['precision', 'recall']
        """
        filtered_metrics = list()
        for metric, phase_list, tags_list, micro_macro_list in self.logged_metrics:
            add_metric_1 = (
                True
                if tag_group is None
                else all([tag in tags_list for tag in tag_group])
            )
            add_metric_2 = (
                True
                if phase_group is None
                else all([phase in phase_list for phase in phase_group])
            )
            add_metric_3 = (
                True
                if micro_macro_group is None
                else all(
                    [
                        micro_macro in micro_macro_list
                        for micro_macro in micro_macro_group
                    ]
                )
            )
            add_metric_4 = True if exclude is None else metric not in exclude

            if add_metric_1 and add_metric_2 and add_metric_3 and add_metric_4:
                filtered_metrics.append(metric)

        return list(set(filtered_metrics))

    @classmethod
    def as_flat_list(cls):
        """
        return logged_metrics as flat list
        ----------------------------------
        :return: [list] of [str], e.g. ['all_precision_micro', 'all_precision_macro', 'fil_precision_micro', ..]
        """
        flat_list = list()
        for metric, _, tags_list, micro_macro_list in cls.logged_metrics:
            for tags in tags_list:
                if micro_macro_list == ["simple"]:
                    flat_list.append(f"{tags}_{metric}")
                else:
                    for style in micro_macro_list:
                        flat_list.append(f"{tags}_{metric}_{style}")
        return flat_list



import numpy as np
import pandas as pd
import pytest
from pkg_resources import resource_filename
from nerblackbox.modules.ner_training.metrics.ner_metrics import (
    NerMetrics,
    get_rid_of_special_tokens,
    convert_to_chunk,
)
from typing import List


########################################################################################################################
########################################################################################################################
########################################################################################################################
class TestNerMetricsTable:

    metrics_simple = ["acc"]
    metrics_micro_macro = ["precision", "recall", "f1"]
    metrics = metrics_simple + metrics_micro_macro

    df = {
        level: pd.read_csv(
            resource_filename(
                "nerblackbox", f"tests/test_data/test_ner_metrics_{level}.csv"
            ),
            sep=";",
        )
        for level in ["token", "chunk"]
    }
    level: str
    labels: List[str]
    micro: bool
    macro: bool

    def test_predictions_from_csv_token(self):
        self.level = "token"
        self.labels = ["all", "fil", "A", "B", "C", "O"]
        self.micro = True
        self.macro = True
        self._test_predictions_from_csv()

    def test_predictions_from_csv_chunk(self):
        self.level = "chunk"
        self.labels = ["fil"]
        self.micro = True
        self.macro = False
        self._test_predictions_from_csv()

    ####################################################################################################################
    # TEST #############################################################################################################
    ####################################################################################################################
    def _test_predictions_from_csv(self) -> None:
        """
        test true against pred values for all rows and labels
        """
        nr_rows = len(self.df[self.level])
        true = self._seq2array(self.df[self.level]["sequence"].iloc[0])

        for row in range(nr_rows):
            pred = self._seq2array(self.df[self.level]["sequence"].iloc[row])
            tested_columns = list()
            for labels in self.labels:
                _tested_columns = self._single_row_and_label_category_test(
                    true, pred, row=row, labels=labels
                )
                tested_columns.extend(_tested_columns)

            assert set(tested_columns) == set(self.df[self.level].columns[2:])

    def _single_row_and_label_category_test(
        self, true: np.array, pred: np.array, row: int, labels: str
    ) -> List[str]:
        """
        test true against pred values for single row in csv and specific labels

        Args:
            true:   true values, e.g. [A, A, O, O, B]
            pred:   pred values, e.g. [A, A, A, O, B]
            row:    e.g. 2
            labels: e.g. 'all', 'A', 'B'

        Returns:
            tested_metrics: labels_metrics, e.g. ['all-precision', 'B-recall', ..]
        """

        def get_tag_list(_labels):
            if labels == "all":
                return None
            elif labels == "fil":
                return ["A", "B"]
            else:
                return _labels

        ner_metrics = NerMetrics(
            true,
            pred,
            tag_list=get_tag_list(labels),
            level=self.level,
            plain_tags=True,
            verbose=True,
        )
        ner_metrics.compute(self.metrics)
        ner_metrics_results = ner_metrics.results_as_dict()

        metrics_extended = self._extend_all_metrics(labels)
        _tested_metrics = list()
        for metric in metrics_extended:
            labels_metric = f"{labels}-{metric}"

            ner_metrics_results_metric = ner_metrics_results[
                self._extend_metric(labels, metric)
            ]
            assert ner_metrics_results_metric == self._pytest_approx(
                self.df[self.level][labels_metric][row]
            ), f"pred_{row}, {labels_metric}"

            _tested_metrics.append(labels_metric)

        return _tested_metrics

    ####################################################################################################################
    # HELPERS ##########################################################################################################
    ####################################################################################################################
    def _extend_all_metrics(self, _labels):
        """
        derive _metrics_extended from self.metrics in case of labels == 'fil', 'all'
        ----------------------------------------------------------------------------
        :param _labels: [str], e.g. 'all', 'A', 'B'
        :return: _metrics_extended [list] of [str], e.g. ['acc', 'precision_micro', 'precision_macro', ..]
        """
        _metrics_extended = list()
        if _labels == "all":
            for field in self.metrics_simple:
                if field in self.metrics:
                    _metrics_extended.append(field)
            for field in self.metrics_micro_macro:
                if field in self.metrics:
                    if self.macro:
                        _metrics_extended.append(f"{field}_macro")
        elif _labels == "fil":
            for field in self.metrics_micro_macro:
                if field in self.metrics:
                    if self.macro:
                        _metrics_extended.append(f"{field}_macro")
                    if self.micro:
                        _metrics_extended.append(f"{field}_micro")
        else:
            _metrics_extended = self.metrics_micro_macro
        return _metrics_extended

    def _extend_metric(self, _labels, _metric):
        """
        extend _metric with '_micro' suffix
        -----------------------------------
        :param _labels: [str], e.g. 'all', 'A', 'B'
        :param _metric: [str], e.g. 'acc', 'precision', 'recall', 'f1', ..
        :return: _extended_metric: [str], e.g. 'precision_micro'
        """
        return (
            _metric
            if (_labels in ["fil", "all"] or _metric in self.metrics_simple)
            else f"{_metric}_micro"
        )

    @staticmethod
    def _seq2array(_str):
        """
        convert sequence str to numpy array
        -----------------------------------
        :param _str: sequence str [str],      e.g. '[A, A, O, O, B]'
        :return:                  [np array], e.g. np.array(['A', 'A', 'O', 'O', 'B'])
        """
        return np.array(_str.strip("[").strip("]").replace(" ", "").split(","))

    @staticmethod
    def _pytest_approx(number):
        """
        get acceptable pytest range for number
        --------------------------------------
        :param number: [float], e.g. 0.82
        :return: pytest range, e.g. 0.82 +- 0.01
        """
        test_precision = 0.01
        return pytest.approx(number, abs=test_precision)


########################################################################################################################
########################################################################################################################
########################################################################################################################
class TestNerMetrics:
    @pytest.mark.parametrize(
        "input_sequence, " "output_sequence",
        [
            (
                ["A", "A", "O", "O", "B"],
                ["A", "A", "O", "O", "B"],
            ),
            (
                ["[CLS]", "A", "A", "O", "O", "B", "[SEP]"],
                ["O", "A", "A", "O", "O", "B", "O"],
            ),
            (
                ["[CLS]", "A", "A", "[UNK]", "O", "O", "B", "[SEP]"],
                ["O", "A", "A", "O", "O", "O", "B", "O"],
            ),
        ],
    )
    def test_get_rid_of_special_tokens(
        self, input_sequence: List[str], output_sequence: List[str]
    ):
        test_output_sequence = get_rid_of_special_tokens(input_sequence)
        assert (
            test_output_sequence == output_sequence
        ), f"{test_output_sequence} != {output_sequence}"

    @pytest.mark.parametrize(
        "input_sequence, " "to_bio, " "output_sequence",
        [
            (
                ["[CLS]", "A", "A", "[UNK]", "O", "O", "B", "[SEP]"],
                True,
                ["O", "B-A", "I-A", "O", "O", "O", "B-B", "O"],
            ),
            (
                ["[CLS]", "B-A", "I-A", "[UNK]", "O", "O", "B-B", "[SEP]"],
                False,
                ["O", "B-A", "I-A", "O", "O", "O", "B-B", "O"],
            ),
            (
                ["[CLS]", "A", "A", "[UNK]", "O", "O", "B", "[SEP]"],
                False,
                None,
            ),
            (
                ["[CLS]", "B-A", "I-A", "[UNK]", "O", "O", "B-B", "[SEP]"],
                True,
                None,
            ),
        ],
    )
    def test_convert_to_chunk(
        self, input_sequence: List[str], to_bio: bool, output_sequence: List[str]
    ):
        if output_sequence is not None:
            test_output_sequence = convert_to_chunk(input_sequence, to_bio)
            assert (
                test_output_sequence == output_sequence
            ), f"{test_output_sequence} != {output_sequence}"
        else:
            with pytest.raises(Exception):
                convert_to_chunk(input_sequence, to_bio)


if __name__ == "__main__":
    test_ner_metrics_table = TestNerMetricsTable()
    test_ner_metrics_table.test_predictions_from_csv_token()
    test_ner_metrics_table.test_predictions_from_csv_chunk()

    test_ner_metrics = TestNerMetrics()
    test_ner_metrics.test_get_rid_of_special_tokens()
    test_ner_metrics.test_convert_to_chunk()



from nerblackbox.modules.datasets.formatter.swedish_ner_corpus_formatter import (
    SwedishNerCorpusFormatter,
)
from nerblackbox.modules.datasets.formatter.sic_formatter import SICFormatter
from nerblackbox.modules.datasets.formatter.suc_formatter import SUCFormatter
from nerblackbox.modules.datasets.formatter.conll2003_formatter import (
    CoNLL2003Formatter,
)
from nerblackbox.modules.datasets.formatter.base_formatter import BaseFormatter


class AutoFormatter:
    @staticmethod
    def for_dataset(ner_dataset: str) -> BaseFormatter:
        if ner_dataset == "swedish_ner_corpus":
            return SwedishNerCorpusFormatter()
        elif ner_dataset == "conll2003":
            return CoNLL2003Formatter()
        elif ner_dataset == "sic":
            return SICFormatter()
        elif ner_dataset == "suc":
            return SUCFormatter()
        else:
            raise Exception(f"ner_dataset = {ner_dataset} unknown.")



import os
from os.path import join, isfile, isdir
import glob
import mlflow
import shutil
from argparse import Namespace
from pkg_resources import Requirement
from pkg_resources import resource_filename, resource_isdir
import pandas as pd
from mlflow.tracking import MlflowClient

from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.utils.util_functions import epoch2checkpoint
from nerblackbox.modules.utils.util_functions import (
    get_run_name,
    compute_mean_and_dmean,
)
from nerblackbox.modules.experiment_results import ExperimentResults
from typing import Optional, Any, Tuple, Union, Dict, List
from pandas import DataFrame

DATASETS_DOWNLOAD = [
    "conll2003",
    "swedish_ner_corpus",
    "sic",
]


class NerBlackBoxMain:
    def __init__(
        self,
        flag: str,
        usage: str = "cli",
        dataset_name: Optional[str] = None,  # analyze_data & set_up_dataset
        modify: bool = True,  # set_up_dataset
        val_fraction: float = 0.3,  # set_up_dataset
        verbose: bool = False,
        experiment_name: Optional[str] = None,
        run_name: Optional[str] = None,  # run_experiment
        device: Any = "gpu",  # run_experiment
        fp16: bool = False,  # run_experiment
        text_input: Optional[str] = None,  # predict
        ids: Tuple[str, ...] = (),  # get_experiments, get_experiments_results
        as_df: bool = True,  # get_experiments, get_experiments_results
        results: bool = False,  # clear_data
    ):
        """
        :param flag:            [str], e.g. 'analyze_data', 'set_up_dataset', 'run_experiment', ..
        :param usage:           [str] 'cli' or 'api'
        :param dataset_name     [str] e.g. 'swedish_ner_corpus'
        :param modify           [bool] if True: modify tags as specified in method modify_ner_tag_mapping()
        :param val_fraction     [float] e.g. 0.3
        :param verbose          [bool]
        :param experiment_name: [str], e.g. 'exp0'
        :param run_name:        [str or None], e.g. 'runA'
        :param device:          [torch device]
        :param fp16:            [bool]
        :param text_input:      [str], e.g. 'this is some text that needs to be annotated'
        :param ids:             [tuple of str], experiment_ids to include
        :param as_df:           [bool] if True, return pandas DataFrame, else return dict
        :param results:         [bool] if True, clear not only checkpoints but also mlflow, tensorboard and logs
        """
        self._assert_flag(flag)

        os.environ["MLFLOW_TRACKING_URI"] = env_variable("DIR_MLFLOW")

        self.flag = flag
        self.usage = usage
        self.dataset_name = dataset_name  # analyze_data & set_up_dataset
        self.modify = modify  # set_up_dataset
        self.val_fraction = val_fraction  # set_up_dataset
        self.verbose = verbose
        self.experiment_name = experiment_name
        self.run_name = run_name  # run_experiment
        self.device = device  # run_experiment
        self.fp16 = fp16  # run_experiment
        self.text_input = text_input  # predict
        self.ids = ids  # get_experiments, get_experiments_results
        self.as_df = as_df  # get_experiments, get_experiments_results
        self.results = results  # clear_data

        data_dir = env_variable("DATA_DIR")
        if os.path.isdir(data_dir):
            self._set_client_and_get_experiments()
        else:
            # will be set in init() method
            self.client = None
            self.experiment_id2name = None
            self.experiment_name2id = None

    ####################################################################################################################
    # MAIN #############################################################################################################
    ####################################################################################################################
    def main(self):

        ################################################################################################################
        # init
        ################################################################################################################
        if self.flag == "init":
            self._create_data_directory()
            self._set_client_and_get_experiments()

        ################################################################################################################
        # download
        ################################################################################################################
        elif self.flag == "download":
            for _dataset_name in DATASETS_DOWNLOAD:
                self.set_up_dataset(_dataset_name)

        ################################################################################################################
        # analyze_data
        ################################################################################################################
        elif self.flag == "analyze_data":
            self._assert_flag_arg("dataset_name")
            self.analyze_data()

        ################################################################################################################
        # set_up_dataset
        ################################################################################################################
        elif self.flag == "set_up_dataset":
            self._assert_flag_arg("dataset_name")
            self.set_up_dataset(self.dataset_name)

        ################################################################################################################
        # show_experiment_configs
        ################################################################################################################
        elif self.flag == "show_experiment_configs":
            self.show_experiment_configs()

        ################################################################################################################
        # show_experiment_config
        ################################################################################################################
        elif self.flag == "show_experiment_config":
            self._assert_flag_arg("experiment_name")
            self.show_experiment_config()

        ################################################################################################################
        # run_experiment
        ################################################################################################################
        elif self.flag == "run_experiment":
            self._assert_flag_arg("experiment_name")
            self.run_experiment()

        ################################################################################################################
        # get_experiment_results
        ################################################################################################################
        elif self.flag == "get_experiment_results":
            self._assert_flag_arg("experiment_name")
            self._assert_usage()
            return self.get_experiment_results()

        ################################################################################################################
        # get_experiments
        ################################################################################################################
        elif self.flag == "get_experiments":
            self._assert_usage()
            return self.get_experiments()

        ################################################################################################################
        # get_experiments_best_runs
        ################################################################################################################
        elif self.flag == "get_experiments_results":
            self._assert_usage()
            return self.get_experiments_results()

        ################################################################################################################
        # predict
        ################################################################################################################
        elif self.flag == "predict":
            self._assert_flag_arg("experiment_name")
            self._assert_flag_arg("text_input")
            return self.predict()

        ################################################################################################################
        # predict
        ################################################################################################################
        elif self.flag == "predict_proba":
            self._assert_flag_arg("experiment_name")
            self._assert_flag_arg("text_input")
            return self.predict_proba()

        ################################################################################################################
        # clear
        ################################################################################################################
        elif self.flag == "clear_data":
            return self.clear_data()

    ####################################################################################################################
    # FLAGS ############################################################################################################
    ####################################################################################################################
    def analyze_data(self) -> None:
        """
        :used attr: dataset_name [str] e.g. 'swedish_ner_corpus'
        :used attr: verbose      [bool]
        """
        _parameters = {
            "ner_dataset": self.dataset_name,
            "verbose": self.verbose,
        }

        mlflow.projects.run(
            uri=resource_filename(Requirement.parse("nerblackbox"), "nerblackbox"),
            entry_point="analyze_data",
            experiment_name="Default",
            parameters=_parameters,
            use_conda=False,
        )

    def clear_data(self) -> None:
        """
        :used attr: clear_all [bool] if True, clear not only checkpoints but also mlflow, tensorboard and logs
        """
        data_dir = env_variable("DATA_DIR")
        results_dir = join(data_dir, "results")
        assert isdir(results_dir), f"directory {results_dir} does not exist."

        # checkpoints
        objects_to_remove = glob.glob(join(results_dir, "checkpoints", "*"))  # list

        # results (mlflow, tensorboard, ..)
        if self.results:
            results_files = (
                glob.glob(join(results_dir, "mlruns", "*"))
                + glob.glob(join(results_dir, "mlruns", ".*"))
                + glob.glob(join(results_dir, "tensorboard", "*"))
                + glob.glob(join(results_dir, "logs.log"))
                + glob.glob(join(results_dir, "*.npy"))
            )
            objects_to_remove.extend(results_files)

        if len(objects_to_remove) == 0:
            print(f"There is no data to remove in {results_dir}")
        else:
            for elem in objects_to_remove:
                print(elem)
            while 1:
                answer = input("Do you want to remove the above files? (y/n) ")
                if answer == "y":
                    for elem in objects_to_remove:
                        if isfile(elem):
                            os.remove(elem)
                        elif isdir(elem):
                            shutil.rmtree(elem, ignore_errors=False)
                        else:
                            raise ValueError(
                                f"object {elem} is neither a file nor a dir and cannot be removed"
                            )
                    print(f"Files removed")
                    break
                elif answer == "n":
                    print(f"No files removed")
                    break
                else:
                    print("Please enter either y or n")

    def get_experiment_results(self) -> Optional[ExperimentResults]:
        """
        :used attr: experiment_name [str], e.g. 'exp0'
        :return: experiment_results [ExperimentResults]
        """
        from nerblackbox.modules.ner_training.ner_model_predict import (
            NerModelPredict,
        )

        assert (
            self.experiment_id2name is not None
        ), f"ERROR! self.experiment_id2name is None."
        if self.experiment_name not in self.experiment_name2id.keys():
            print(f"no experiment with experiment_name = {self.experiment_name} found")
            print(f"experiments that were found:")
            print(self.experiment_name2id)
            return ExperimentResults()

        experiment_id = self.experiment_name2id[self.experiment_name]
        experiment_results = self._get_single_experiment_results(experiment_id)

        if self.usage == "cli":
            print("### single runs ###")
            print(experiment_results.single_runs)
            print()
            print("### average runs ###")
            print(experiment_results.average_runs)
            return None
        else:
            if experiment_results.best_single_run["checkpoint"] is not None:
                best_model = NerModelPredict.load_from_checkpoint(
                    experiment_results.best_single_run["checkpoint"]
                )
                experiment_results._set_best_model(best_model)

            return experiment_results

    def get_experiments(self) -> Optional[Union[DataFrame, Dict]]:
        r"""
        :used attr: ids    [tuple] of [str], e.g. ('4', '5')
        :used attr: as_df  [bool] if True, return [pandas DataFrame], else return [dict]
        :return: experiments_overview [pandas DataFrame] or [dict]
        """
        assert self.ids is not None, f"ERROR! self.ids is None."
        experiments_id2name_filtered = self._filter_experiments_by_ids(self.ids)

        experiments_overview = sorted(
            [
                {"experiment_id": k, "experiment_name": v}
                for k, v in experiments_id2name_filtered.items()
            ],
            key=lambda elem: elem["experiment_id"],
        )
        df = pd.DataFrame(experiments_overview) if self.as_df else experiments_overview
        if self.usage == "cli":
            print("### experiments ###")
            print(df)
            return None
        else:
            return df

    def get_experiments_results(self) -> Optional[Namespace]:
        r"""
        :used attr: ids     [tuple] of [str], e.g. ('4', '5')
        :used attr: as_df   [bool] if True, return [pandas DataFrame], else return [dict]
        :used attr: verbose [bool]
        :return: experiments_results: [Namespace] w/ attributes = 'best_single_runs', 'best_average_runs'
                                                   & values = [pandas DataFrame] or [dict]
        """
        assert self.ids is not None, f"ERROR! self.ids is None."
        experiments_filtered = self._filter_experiments_by_ids(self.ids)

        best_single_runs_overview = list()
        best_average_runs_overview = list()
        for _id in sorted(list(experiments_filtered.keys())):
            experiment_results = self._get_single_experiment_results(
                _id,
            )
            if experiment_results.best_single_run:
                best_single_runs_overview.append(experiment_results.best_single_run)
            if experiment_results.best_average_run:
                best_average_runs_overview.append(experiment_results.best_average_run)

        df_single = (
            pd.DataFrame(best_single_runs_overview)
            if self.as_df
            else best_single_runs_overview
        )
        df_average = (
            pd.DataFrame(best_average_runs_overview)
            if self.as_df
            else best_average_runs_overview
        )
        if self.usage == "cli":
            print("### best single runs ###")
            print(df_single)
            print()
            print("### best average runs ###")
            print(df_average)
            return None
        else:
            return Namespace(
                **{
                    "best_single_runs": df_single,
                    "best_average_runs": df_average,
                }
            )

    def predict(self) -> Optional[List[Namespace]]:
        """
        :used attr: experiment_name [str], e.g. 'exp1'
        :used attr: text_input      [str], e.g. 'this is some text that needs to be annotated'
        :return: predictions [list] of [Namespace] with .internal [list] of (word, tag) tuples
                                                   and  .external [list] of (word, tag) tuples
        """
        nerbb = NerBlackBoxMain(
            "get_experiment_results", experiment_name=self.experiment_name, usage="api"
        )
        experiment_results = nerbb.main()
        predictions = experiment_results.best_model.predict(self.text_input)
        if self.usage == "cli":
            print(predictions[0].external)
            return None
        else:
            return predictions

    def predict_proba(self) -> Optional[List[Namespace]]:
        """
        :used attr: experiment_name [str], e.g. 'exp1'
        :used attr: text_input      [str], e.g. 'this is some text that needs to be annotated'
        :return: predictions [list] of [Namespace] with .internal [list] of (word, tag) tuples
                                                   and  .external [list] of (word, tag) tuples
        """
        nerbb = NerBlackBoxMain(
            "get_experiment_results", experiment_name=self.experiment_name, usage="api"
        )
        experiment_results = nerbb.main()
        predictions = experiment_results.best_model.predict_proba(self.text_input)
        if self.usage == "cli":
            print(predictions[0].external)
            return None
        else:
            return predictions

    def run_experiment(self) -> None:
        """
        :used attr: experiment_name [str],         e.g. 'exp1'
        :used attr: run_name        [str] or None, e.g. 'runA'
        :used attr: device          [torch device]
        :used attr: fp16            [bool]
        """
        _parameters = {
            "experiment_name": self.experiment_name,
            "run_name": self.run_name if self.run_name else "",
            "device": self.device,
            "fp16": int(self.fp16),
        }

        mlflow.projects.run(
            uri=resource_filename(Requirement.parse("nerblackbox"), "nerblackbox"),
            entry_point="run_experiment",
            experiment_name=self.experiment_name,
            parameters=_parameters,
            use_conda=False,
        )

        self._get_experiments()  # needs to updated to get results from experiment that was run
        self.get_experiment_results()

    def set_up_dataset(self, _dataset_name: str) -> None:
        """
        :param _dataset_name:    [str] e.g. 'swedish_ner_corpus'
        :used attr: modify       [bool] if True: modify tags as specified in method modify_ner_tag_mapping()
        :used attr: val_fraction [float] e.g. 0.3
        :used attr: verbose      [bool]
        """

        _parameters = {
            "ner_dataset": _dataset_name,
            "modify": self.modify,
            "val_fraction": self.val_fraction,
            "verbose": self.verbose,
        }

        mlflow.projects.run(
            uri=resource_filename(Requirement.parse("nerblackbox"), "nerblackbox"),
            entry_point="set_up_dataset",
            experiment_name="Default",
            parameters=_parameters,
            use_conda=False,
        )

    def show_experiment_config(self) -> None:
        """
        print experiment config
        -----------------------
        :used attr: experiment_name: [str], e.g. 'exp0'
        """
        from nerblackbox.modules.utils.env_variable import env_variable

        path_experiment_config = join(
            env_variable("DIR_EXPERIMENT_CONFIGS"), f"{self.experiment_name}.ini"
        )
        with open(path_experiment_config, "r") as file:
            lines = file.read()

        print(f"> experiment_config = {path_experiment_config}")
        print()
        print(lines)

    @staticmethod
    def show_experiment_configs() -> None:
        experiment_configs = glob.glob(
            join(env_variable("DIR_EXPERIMENT_CONFIGS"), "*.ini")
        )
        experiment_configs = [
            elem.split("/")[-1].strip(".ini") for elem in experiment_configs
        ]
        experiment_configs = [elem for elem in experiment_configs if elem != "default"]
        for experiment_config in experiment_configs:
            print(experiment_config)

    ####################################################################################################################
    # HELPER ###########################################################################################################
    ####################################################################################################################
    @staticmethod
    def _create_data_directory() -> None:
        if resource_isdir(Requirement.parse("nerblackbox"), "nerblackbox/modules/data"):
            data_source = resource_filename(
                Requirement.parse("nerblackbox"), "nerblackbox/modules/data"
            )

            data_dir = env_variable("DATA_DIR")
            print("data_source =", data_source)
            print("data_target =", data_dir)
            if os.path.isdir(data_dir):
                print(f"init: target {data_dir} already exists")
            else:
                shutil.copytree(data_source, data_dir)
                print(f"init: target {data_dir} created")
        else:
            print("init not executed successfully")
            exit(0)

    def _set_client_and_get_experiments(self) -> None:
        """
        :created attr: client             [Mlflow client]
        :created attr: experiment_id2name [dict] w/ keys = experiment_id [str] & values = experiment_name [str]
        :created attr: experiment_name2id [dict] w/ keys = experiment_name [str] & values = experiment_id [str]
        :return: -
        """
        self.client = MlflowClient()
        self._get_experiments()

    def _get_experiments(self) -> None:
        """
        :created attr: experiment_id2name [dict] w/ keys = experiment_id [str] & values = experiment_name [str]
        :created attr: experiment_name2id [dict] w/ keys = experiment_name [str] & values = experiment_id [str]
        :return: -
        """
        assert self.client is not None, f"ERROR! self.client is None."
        self.experiment_id2name = {
            elem["_experiment_id"]: elem["_name"]
            for elem in [
                vars(experiment) for experiment in self.client.list_experiments()
            ]
            if elem["_name"] != "Default"
        }

        assert (
            self.experiment_id2name is not None
        ), f"ERROR! self.experiment_id2name is None."
        self.experiment_name2id = {v: k for k, v in self.experiment_id2name.items()}

    ####################################################################################################################
    # HELPER: SINGLE EXPERIMENT
    ####################################################################################################################
    def _get_single_experiment_results(self, experiment_id: str) -> ExperimentResults:
        r"""
        :param experiment_id: [str], e.g. '0'
        :return: experiment_results: [ExperimentResults]
        """
        assert (
            self.experiment_id2name is not None
        ), f"ERROR! self.experiment_id2name is None."

        experiment_name = self.experiment_id2name[experiment_id]
        runs = self.client.search_runs(experiment_id)

        _experiment, _single_runs, _average_runs = self._parse_and_create_dataframe(
            runs,
        )

        # best run
        if _experiment is not None and _single_runs is not None:
            _df_best_single_run = _single_runs.iloc[0, :]
            best_single_run_id = _df_best_single_run[("info", "run_id")]
            best_single_run_name_nr = _df_best_single_run[("info", "run_name_nr")]
            best_single_run_epoch_best = _df_best_single_run[("metrics", "epoch_best")]
            best_single_run_epoch_best_val_chk_f1_micro = _df_best_single_run[
                ("metrics", "epoch_best_val_chk_f1_micro")
            ]
            best_single_run_epoch_best_test_chk_f1_micro = _df_best_single_run[
                ("metrics", "epoch_best_test_chk_f1_micro")
            ]

            checkpoint = join(
                env_variable("DIR_CHECKPOINTS"),
                experiment_name,
                best_single_run_name_nr,
                epoch2checkpoint(best_single_run_epoch_best),
            )

            _best_single_run = {
                "experiment_id": experiment_id,
                "experiment_name": experiment_name,
                "run_id": best_single_run_id,
                "run_name_nr": best_single_run_name_nr,
                "epoch_best_val_chk_f1_micro": best_single_run_epoch_best_val_chk_f1_micro,
                "epoch_best_test_chk_f1_micro": best_single_run_epoch_best_test_chk_f1_micro,
                "checkpoint": checkpoint if isfile(checkpoint) else None,
            }
        else:
            _best_single_run = dict()

        # best run average
        if _experiment is not None and _average_runs is not None:
            _df_best_average_run = _average_runs.iloc[0, :]
            best_average_run_name = _df_best_average_run[("info", "run_name")]
            best_average_run_epoch_best_val_chk_f1_micro = _df_best_average_run[
                ("metrics", "epoch_best_val_chk_f1_micro")
            ]
            d_best_average_run_epoch_best_val_chk_f1_micro = _df_best_average_run[
                ("metrics", "d_epoch_best_val_chk_f1_micro")
            ]
            best_average_run_epoch_best_test_chk_f1_micro = _df_best_average_run[
                ("metrics", "epoch_best_test_chk_f1_micro")
            ]
            d_best_average_run_epoch_best_test_chk_f1_micro = _df_best_average_run[
                ("metrics", "d_epoch_best_test_chk_f1_micro")
            ]

            _best_average_run = {
                "experiment_id": experiment_id,
                "experiment_name": experiment_name,
                "run_name": best_average_run_name,
                "epoch_best_val_chk_f1_micro": best_average_run_epoch_best_val_chk_f1_micro,
                "d_epoch_best_val_chk_f1_micro": d_best_average_run_epoch_best_val_chk_f1_micro,
                "epoch_best_test_chk_f1_micro": best_average_run_epoch_best_test_chk_f1_micro,
                "d_epoch_best_test_chk_f1_micro": d_best_average_run_epoch_best_test_chk_f1_micro,
            }
        else:
            _best_average_run = dict()
        return ExperimentResults(
            _experiment,
            _single_runs,
            _average_runs,
            _best_single_run,
            _best_average_run,
        )

    ####################################################################################################################
    # HELPER: ALL EXPERIMENTS
    ####################################################################################################################
    def _filter_experiments_by_ids(self, _ids: Tuple[str, ...]) -> Dict:
        r"""
        get _experiments_id2name [dict] with _ids as keys
        -------------------------------------------------
        :param _ids:  [tuple] of [str], e.g. ('4', '5')
        :return: _experiments_id2name [dict] w/ keys = experiment_id [str] & values = experiment_name [str]
        """
        assert (
            self.experiment_id2name is not None
        ), f"ERROR! self.experiment_id2name is None."

        if len(_ids) == 0:
            _experiments_id2name = self.experiment_id2name
        else:
            _experiments_id2name = {
                k: v for k, v in self.experiment_id2name.items() if k in _ids
            }
        return _experiments_id2name

    @staticmethod
    def _parse_and_create_dataframe(
        _runs: List,
    ) -> Tuple[DataFrame, DataFrame, DataFrame]:
        r"""
        turn mlflow Run objects (= search_runs() results) into data frames
        ------------------------------------------------------------------
        :param _runs:   [list] of [mlflow.entities.Run objects]
        :return: _experiment:   [pandas DataFrame] overview on experiment parameters
        :return: _single_runs:  [pandas DataFrame] overview on single  run parameters & results
        :return: _average_runs: [pandas DataFrame] overview on average run parameters & results
        """
        fields_metrics = [
            "epoch_best",
            "epoch_stopped",
            "epoch_best_val_chk_f1_micro",
            "epoch_best_test_chk_f1_micro",
        ]

        ###########################################
        # parameters_experiment & parameters_runs
        ###########################################
        parameters_runs: Dict[Tuple, Any] = dict()
        parameters_experiment: Dict[Any, Any] = dict()
        for i in range(len(_runs)):
            if len(_runs[i].data.metrics) == 0:  # experiment
                parameters_experiment = {
                    k: [v] for k, v in _runs[i].data.params.items()
                }
            else:  # run
                if ("info", "run_id") not in parameters_runs.keys():
                    parameters_runs[("info", "run_id")] = [_runs[i].info.run_id]
                else:
                    parameters_runs[("info", "run_id")].append(_runs[i].info.run_id)

                if ("info", "run_name_nr") not in parameters_runs.keys():
                    parameters_runs[("info", "run_name_nr")] = [
                        _runs[i].data.tags["mlflow.runName"]
                    ]
                else:
                    parameters_runs[("info", "run_name_nr")].append(
                        _runs[i].data.tags["mlflow.runName"]
                    )

                for k, v in _runs[i].data.params.items():
                    if ("params", k) not in parameters_runs.keys():
                        parameters_runs[("params", k)] = [v]
                    else:
                        parameters_runs[("params", k)].append(v)

                for k in fields_metrics:
                    if ("metrics", k) not in parameters_runs.keys():
                        try:
                            parameters_runs[("metrics", k)] = [_runs[i].data.metrics[k]]
                        except:
                            parameters_runs[("metrics", k)] = [-1]
                    else:
                        try:
                            parameters_runs[("metrics", k)].append(
                                _runs[i].data.metrics[k]
                            )
                        except:
                            parameters_runs[("metrics", k)] = [-1]

        _experiment = pd.DataFrame(parameters_experiment, index=["experiment"]).T
        for k in ["epoch_best", "epoch_stopped"]:
            try:
                parameters_runs[("metrics", k)] = [
                    int(elem) for elem in parameters_runs[("metrics", k)]
                ]
            except:
                parameters_runs[("metrics", k)] = [-1]

        ###########################################
        # parameters_runs_average
        ###########################################
        def average(_parameters_runs):
            _parameters_runs_average = {("info", "run_name"): list()}
            _parameters_runs_average.update(
                {
                    k: list()
                    for k in _parameters_runs.keys()
                    if k[0] not in ["info", "metrics"]
                }
            )
            _parameters_runs_average.update(
                {
                    k: list()
                    for k in [
                        ("metrics", "epoch_best_val_chk_f1_micro"),
                        ("metrics", "d_epoch_best_val_chk_f1_micro"),
                        ("metrics", "epoch_best_test_chk_f1_micro"),
                        ("metrics", "d_epoch_best_test_chk_f1_micro"),
                    ]
                }
            )

            runs_name_nr = _parameters_runs[("info", "run_name_nr")]
            nr_runs = len(runs_name_nr)
            runs_name = list(
                set([get_run_name(run_name_nr) for run_name_nr in runs_name_nr])
            )
            indices = {
                run_name: [
                    idx
                    for idx in range(nr_runs)
                    if get_run_name(runs_name_nr[idx]) == run_name
                ]
                for run_name in runs_name
            }

            def get_mean_and_dmean(_parameters_runs, phase):
                values = [
                    _parameters_runs[("metrics", f"epoch_best_{phase}_chk_f1_micro")][
                        idx
                    ]
                    for idx in indices[run_name]
                ]
                return compute_mean_and_dmean(values)

            #######################
            # loop over runs_name
            #######################
            for run_name in runs_name:
                # add ('info', 'run_name')
                _parameters_runs_average[("info", "run_name")].append(run_name)

                # add ('params', *)
                random_index = indices[run_name][0]
                keys_kept = [k for k in _parameters_runs.keys() if k[0] == "params"]
                for key in keys_kept:
                    _parameters_runs_average[key].append(
                        _parameters_runs[key][random_index]
                    )

                # add ('metrics', *)
                val_mean, val_dmean = get_mean_and_dmean(_parameters_runs, "val")
                test_mean, test_dmean = get_mean_and_dmean(_parameters_runs, "test")
                metrics = {
                    "epoch_best_val_chk_f1_micro": val_mean,
                    "d_epoch_best_val_chk_f1_micro": val_dmean,
                    "epoch_best_test_chk_f1_micro": test_mean,
                    "d_epoch_best_test_chk_f1_micro": test_dmean,
                }
                for k in metrics.keys():
                    key = ("metrics", k)
                    _parameters_runs_average[key].append(metrics[k])

            return _parameters_runs_average

        ###########################################
        # sort & return
        ###########################################
        by = ("metrics", "epoch_best_val_chk_f1_micro")
        try:
            _single_runs = pd.DataFrame(parameters_runs).sort_values(
                by=by, ascending=False
            )
        except:
            _single_runs = None

        try:
            parameters_runs_average = average(parameters_runs)
            _average_runs = pd.DataFrame(parameters_runs_average).sort_values(
                by=by, ascending=False
            )
        except:
            _average_runs = None

        return _experiment, _single_runs, _average_runs

    ####################################################################################################################
    # HELPER: ADDITONAL
    ####################################################################################################################
    @staticmethod
    def _assert_flag(flag: str) -> None:
        if flag is None:
            message = f"> missing flag (e.g. init, set_up_dataset, run_experiment)"
            print(message)
            exit(0)

    def _assert_flag_arg(self, flag_arg: str) -> None:
        if getattr(self, flag_arg) is None:
            message = f"> missing argument: nerbb {self.flag} <{flag_arg}>"
            print(message)
            exit(0)

    def _assert_usage(self) -> None:
        assert self.usage in ["cli", "api"], "missing usage"

    @staticmethod
    def show_as_df(_dict: Dict) -> DataFrame:
        return pd.DataFrame(_dict)



import os
from typing import Dict, List
import pandas as pd
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
    InputExample,
)


class CsvReader:
    """
    reads data (tags & text) from csv and
    - creates list of InputExamples
    - gets list of tags
    """

    def __init__(
        self,
        path,
        tokenizer,
        do_lower_case,
        csv_file_separator="\t",
        default_logger=None,
    ):
        """
        :param path:               [str] to folder that contains dataset csv files (train, val, test)
        :param tokenizer:          [transformers Tokenizer]
        :param do_lower_case:      [bool]
        :param csv_file_separator: [str], for datasets' csv files, e.g. '\t'
        """
        # input arguments
        self.path = path
        self.tokenizer = tokenizer
        self.do_lower_case = do_lower_case
        self.csv_file_separator = csv_file_separator
        self.default_logger = default_logger

        # additional attributes
        self.token_count = None

        # data & tag_list
        self.data: Dict[str, pd.DataFrame] = dict()
        self.tag_list_found: List[str] = list()
        for phase in ["train", "val", "test"]:
            # data
            self.data[phase] = self._read_csv(os.path.join(self.path, f"{phase}.csv"))

            # tag list
            tag_list_phase = list(
                set(" ".join(self.data[phase]["tags"].values).split())
            )
            self.tag_list_found = sorted(
                list(set(self.tag_list_found + tag_list_phase))
            )

        self.tag_list: List[str] = ["O"] + [
            elem for elem in self.tag_list_found if elem != "O"
        ]

        if self.default_logger:
            self.default_logger.log_debug(
                f"> tag list found in data: {self.tag_list_found}"
            )
            self.default_logger.log_debug(f"> tag list complete:      {self.tag_list}")

    ####################################################################################################################
    # PUBLIC METHODS
    ####################################################################################################################
    def get_input_examples(self, phase):
        """
        gets list of input examples for specified phase
        -----------------------------------------------
        :param phase: [str], e.g. 'train', 'val', 'test'
        :return: [list] of [InputExample]
        """
        return self._create_list_of_input_examples(self.data[phase], phase)

    ####################################################################################################################
    # PRIVATE METHODS
    ####################################################################################################################
    def _read_csv(self, path: str) -> pd.DataFrame:
        """
        read csv using pandas.

        Note: The csv is expected to
        - have two columns seperated by self.seperator
        - not have a header with column names
        ----------------------------------------------
        :param path: [str]
        :return: [pandas dataframe]
        """
        return pd.read_csv(
            path, names=["tags", "text"], header=None, sep=self.csv_file_separator
        )

    def _create_list_of_input_examples(self, df, set_type):
        """
        create list of input examples from pandas dataframe created from _read_csv() method
        -----------------------------------------------------------------------------------
        :param df:                 [pandas dataframe] with columns 'tags', 'text'
        :param set_type:           [str], e.g. 'train', 'val', 'test'
        :changed attr: token_count [int] total number of tokens in df
        :return: [list] of [InputExample]
        """
        self.token_count = 0

        examples = []
        for i, row in enumerate(df.itertuples()):
            # input_example
            guid = f"{set_type}-{i}"
            text = row.text.lower() if self.do_lower_case else row.text
            tags = row.tags

            input_example = InputExample(guid=guid, text=text, tags=tags)

            # append
            examples.append(input_example)
        return examples



import json

from transformers import AutoModelForTokenClassification

from nerblackbox.modules.ner_training.metrics.logged_metrics import LoggedMetrics
from nerblackbox.modules.ner_training.logging.mlflow_client import MLflowClient
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger
from nerblackbox.modules.ner_training.ner_model import NerModel


class NerModelTrain(NerModel):
    def __init__(self, hparams):
        """
        :param hparams: [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
        """
        super().__init__(hparams)

    ####################################################################################################################
    # Abstract Base Methods ############################################################################################
    ####################################################################################################################
    def _preparations(self):
        """
        :created attr: default_logger    [DefaultLogger]
        :created attr: logged_metrics    [LoggedMetrics]
        :created attr: tokenizer         [transformers AutoTokenizer]
        :created attr: data_preprocessor [DataPreprocessor]
        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
        :created attr: model             [transformers AutoModelForTokenClassification]

        :created attr: mlflow_client          [MLflowClient]
        :created attr: epoch_metrics          [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: classification_reports [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: pretrained_model_name  [str]
        :created attr: dataloader             [dict] w/ keys 'train', 'val', 'test' & values = [torch Dataloader]
        :created attr: optimizer              [torch optimizer]
        :created attr: scheduler              [torch LambdaLR]

        :return: -
        """
        # train/val/test
        self._preparations_train()  # attr: default_logger, logged_metrics, mlflow_client, ..
        self._preparations_data_general()  # attr: tokenizer, data_preprocessor
        self._preparations_data_train()  # attr: tag_list, model, dataloader, optimizer, scheduler

    def _preparations_train(self):
        """
        :created attr: default_logger         [DefaultLogger]
        :created attr: logged_metrics         [LoggedMetrics]
        :created attr: mlflow_client          [MLflowClient]
        :created attr: epoch_metrics          [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: classification_reports [dict] w/ keys = 'val', 'test' & values = [dict]
        :return: -
        """
        self.default_logger = DefaultLogger(
            __file__, log_file=self.log_dirs.log_file, level=self.params.logging_level
        )

        self.logged_metrics = LoggedMetrics()

        self.mlflow_client = MLflowClient(
            experiment_name=self.params.experiment_name,
            run_name=self.params.run_name,
            log_dirs=self.log_dirs,
            logged_metrics=self.logged_metrics.as_flat_list(),
            default_logger=self.default_logger,
        )
        self.mlflow_client.log_params(
            self.params, self.hparams, experiment=self.experiment
        )

        self.epoch_metrics = {"val": dict(), "test": dict()}
        self.classification_reports = {"val": dict(), "test": dict()}

    def _preparations_data_train(self):
        """
        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
        :created attr: model             [transformers AutoModelForTokenClassification]
        :created attr: dataloader        [dict] w/ keys 'train', 'val', 'test' & values = [torch Dataloader]
        :created attr: optimizer         [torch optimizer]
        :created attr: scheduler         [torch LambdaLR]
        :return: -
        """
        # input_examples & tag_list
        input_examples, self.tag_list = self.data_preprocessor.get_input_examples_train(
            prune_ratio={
                "train": self.params.prune_ratio_train,
                "val": self.params.prune_ratio_val,
                "test": self.params.prune_ratio_test,
            },
            dataset_name=self.params.dataset_name,
        )
        self.default_logger.log_debug("> self.tag_list:", self.tag_list)
        self.hparams.tag_list = json.dumps(
            self.tag_list
        )  # save for PREDICT (see below)

        # model
        self.model = AutoModelForTokenClassification.from_pretrained(
            self.pretrained_model_name,
            num_labels=len(self.tag_list),
            return_dict=False,
        )
        self.model.resize_token_embeddings(
            len(self.tokenizer)
        )  # due to additional_special_tokens = NEWLINE_TOKENS

        # dataloader
        self.dataloader = self.data_preprocessor.to_dataloader(
            input_examples, self.tag_list, batch_size=self._hparams.batch_size
        )

        # optimizer
        self.optimizer = self._create_optimizer(
            self._hparams.lr_max, fp16=self.params.fp16
        )

        # learning rate
        self.scheduler = self._create_scheduler(
            self._hparams.lr_warmup_epochs,
            self._hparams.lr_schedule,
            self._hparams.lr_num_cycles,
        )



import os
import numpy as np
from os.path import join
from argparse import Namespace
import pkg_resources

from nerblackbox.modules.utils.env_variable import env_variable


def get_available_datasets():
    """
    get datasets that are available in DIR_DATASETS directory
    ---------------------------------------------------------
    :return: available datasets: [list] of [str], e.g. ['suc', 'swedish_ner_corpus']
    """
    dir_datasets = env_variable("DIR_DATASETS")
    return [
        folder
        for folder in os.listdir(dir_datasets)
        if os.path.isdir(join(dir_datasets, folder))
    ]


def get_dataset_path(dataset):
    """
    get dataset path for dataset
    ----------------------------
    :param dataset:        [str] dataset name, e.g. 'suc', 'swedish_ner_corpus'
    :return: dataset_path: [str] path to dataset directory
    """
    return join(env_variable("DIR_DATASETS"), dataset)


def get_hardcoded_parameters(keys=False):
    """
    :param keys: [bool] whether to return [list] of keys instead of whole [dict]
    :return: _general:  [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    :return: _params:   [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    :return: _hparams:  [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    :return: _log_dirs: [dict] w/ keys = parameter name [str] & values = type [str] --- or [list] of keys
    """
    _general = {
        "experiment_name": "str",
        "run_name": "str",
        "run_name_nr": "str",
        "device": "str",
        "fp16": "bool",
        "experiment_run_name_nr": "str",
    }
    _params = {
        "dataset_name": "str",
        "dataset_tags": "str",
        "prune_ratio_train": "float",
        "prune_ratio_val": "float",
        "prune_ratio_test": "float",
        "pretrained_model_name": "str",
        "uncased": "bool",
        "checkpoints": "bool",
        "logging_level": "str",
        "multiple_runs": "int",
    }
    _hparams = {
        "batch_size": "int",
        "max_seq_length": "int",
        "max_epochs": "int",
        "monitor": "str",
        "min_delta": "float",
        "patience": "int",
        "mode": "str",
        "lr_max": "float",
        "lr_schedule": "str",
        "lr_warmup_epochs": "int",
        "lr_num_cycles": "int",
    }
    _log_dirs = {
        "mlflow": "str",
        "tensorboard": "str",
        "checkpoints": "str",
        "log_file": "str",
        "mlflow_file": "str",
    }
    if keys:
        return (
            list(_general.keys()),
            list(_params.keys()),
            list(_hparams.keys()),
            list(_log_dirs.keys()),
        )
    else:
        return _general, _params, _hparams, _log_dirs


def unify_parameters(_params, _hparams, _log_dirs, _experiment):
    """
    unify parameters (namespaces, bool) to one namespace
    ----------------------------------------------------
    :param _params:             [Namespace] with keys = 'dataset_name', 'dataset_tags', ..
    :param _hparams:            [Namespace] with keys = 'batch_size', 'max_seq_length', ..
    :param _log_dirs:           [Namespace] with keys = 'mlflow', 'tensorboard', ..
    :param _experiment:         [bool]
    :return: _lightning_hparams [Namespace] with keys = all keys from input namespaces + 'experiment'
    """
    _dict = dict()
    _dict.update(vars(_params))
    _dict.update(vars(_hparams))
    _dict.update(vars(_log_dirs))
    _dict.update({"experiment": _experiment})
    _lightning_hparams = Namespace(**_dict)
    _lightning_hparams.device = (
        _lightning_hparams.device.type
    )  # needs to be a string (not torch.device) for logging
    return _lightning_hparams


def split_parameters(_lightning_hparams):
    """
    split namespace to parameters (namespaces, bool)
    ----------------------------------------------------
    :param _lightning_hparams [Namespace] with keys = all keys from output namespaces + 'experiment'
    :return: _params:         [Namespace] with keys = 'dataset_name', 'dataset_tags', ..
    :return: _hparams:        [Namespace] with keys = 'batch_size', 'max_seq_length', ..
    :return: _log_dirs:       [Namespace] with keys = 'mlflow', 'tensorboard', ..
    :return: _experiment:     [bool]
    """
    keys_general, keys_params, keys_hparams, keys_log_dirs = get_hardcoded_parameters(
        keys=True
    )
    _params = Namespace(
        **{
            k: v
            for k, v in vars(_lightning_hparams).items()
            if k in keys_general + keys_params
        }
    )
    _hparams = Namespace(
        **{k: v for k, v in vars(_lightning_hparams).items() if k in keys_hparams}
    )
    _log_dirs = Namespace(
        **{k: v for k, v in vars(_lightning_hparams).items() if k in keys_log_dirs}
    )
    _experiment = vars(_lightning_hparams).get("experiment")
    return _params, _hparams, _log_dirs, _experiment


def get_package_version():
    return pkg_resources.get_distribution("nerblackbox").version


def checkpoint2epoch(_checkpoint_name):
    """
    :param _checkpoint_name: [str], e.g. 'epoch=2.ckpt' or 'epoch=2_v0.ckpt'
    :return: _epoch:         [int], e.g. 2
    """
    return int(_checkpoint_name.split("epoch=")[-1].split("_")[0].replace(".ckpt", ""))


def epoch2checkpoint(_epoch):
    """
    :param _epoch:            [int], e.g. 2
    :return _checkpoint_name: [str], e.g. 'epoch=2.ckpt' or 'epoch=2_v0.ckpt'
    """
    return f"epoch={_epoch}.ckpt"


def get_run_name(_run_name_nr):
    """
    :param _run_name_nr: [str], e.g. 'runA-1'
    :return: _run_name:  [str], e.g. 'runA'
    """
    return _run_name_nr.split("-")[0]


def get_run_name_nr(_run_name, _run_nr):
    """
    :param _run_name:      [str], e.g. 'runA'
    :param _run_nr:        [int], e.g. 1
    :return: _run_name_nr: [str], e.g. 'runA-1'
    """
    return f"{_run_name}-{_run_nr}"


def compute_mean_and_dmean(values):
    """
    compute mean and its error dmean = std deviation / sqrt(N)
    ----------------------------------------------------------
    :param values: [list / np array] of [float]
    :return: mean  [float]
    :return: dmean [float]
    """
    if len(values) == 1:
        return values[0], None
    elif len(values) > 1:
        return np.mean(values), np.std(values) / np.sqrt(len(values))
    else:
        raise Exception(f"cannot compute mean and dmean of empty list!")



from nerblackbox.modules.ner_training.data_preprocessing.tools.bert_dataset import (
    BertDataset,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.csv_reader import (
    CsvReader,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
    InputExample,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_examples_to_tensors import (
    InputExamplesToTensors,
)
from nerblackbox.tests.utils import PseudoDefaultLogger
from nerblackbox.modules.utils.util_functions import get_dataset_path
from torch.utils.data import DataLoader, Sampler, RandomSampler, SequentialSampler
from pkg_resources import resource_filename

from typing import List, Dict, Tuple, Optional, Any

InputExamples = List[InputExample]


class DataPreprocessor:
    def __init__(
        self,
        tokenizer,
        do_lower_case: bool,
        default_logger: PseudoDefaultLogger,
        max_seq_length: int = 64,
    ):
        """
        :param tokenizer:      [transformers Tokenizer]
        :param do_lower_case:  [bool] if True, make text data lowercase
        :param default_logger: [DefaultLogger]
        :param max_seq_length: [int], e.g. 64
        """
        self.tokenizer = tokenizer
        self.do_lower_case = do_lower_case
        self.default_logger = default_logger
        self.max_seq_length = max_seq_length

    def get_input_examples_train(
        self, prune_ratio: Dict[str, float], dataset_name: Optional[str] = None
    ) -> Tuple[Dict[str, InputExamples], List[str]]:
        """
        - get input examples for TRAIN from csv files

        Args:
            dataset_name:     [str], e.g. 'suc'
            prune_ratio:      [dict], e.g. {'train': 1.0, 'val': 1.0, 'test': 1.0} -- pruning ratio for data

        Returns:
            input_examples: [dict] w/ keys = 'train', 'val', 'test' & values = [list] of [InputExample]
            tag_list:       [list] of tags present in the dataset, e.g. ['O', 'PER', ..]
        """
        if dataset_name is None:
            dataset_path = resource_filename("nerblackbox", "tests/test_data")
        else:
            dataset_path = get_dataset_path(dataset_name)

        # csv_reader
        csv_reader = CsvReader(
            dataset_path,
            self.tokenizer,
            do_lower_case=self.do_lower_case,  # can be True (applies .lower()) !!
            default_logger=self.default_logger,
        )

        input_examples = dict()
        for phase in ["train", "val", "test"]:
            # train data
            input_examples_all = csv_reader.get_input_examples(phase)
            input_examples[phase] = self._prune_examples(
                input_examples_all, phase, ratio=prune_ratio[phase]
            )

        return input_examples, self._ensure_completeness_in_case_of_bio_tags(
            csv_reader.tag_list
        )

    def get_input_examples_predict(
        self, examples: List[str]
    ) -> Dict[str, InputExamples]:
        """
        - get input examples for PREDICT from input argument examples

        Args:
            examples:       [list] of [str]

        Returns:
            input_examples: [dict] w/ key = 'predict' & value = [list] of [InputExample]
        """
        # create input_examples
        if self.do_lower_case:
            examples = [example.lower() for example in examples]

        input_examples = {
            "predict": [
                InputExample(
                    guid="",
                    text=example,
                    tags=" ".join(
                        ["O" for _ in range(len(example.split()))]
                    ),  # pseudo tags
                )
                for example in examples
            ]
        }
        # print('TEMP: input_examples:', input_examples)

        return input_examples

    def to_dataloader(
        self,
        input_examples: Dict[str, InputExamples],
        tag_list: List[str],
        batch_size: int,
    ) -> Dict[str, DataLoader]:
        """
        - turn input_examples into dataloader

        Args:
            input_examples: [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
                                      values = [list] of [InputExample]
            tag_list:       [list] of tags present in the dataset, e.g. ['O', 'PER', ..]
            batch_size:     [int]

        Returns:
            _dataloader:    [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
                                      values = [torch Dataloader]
        """
        # input_example_to_tensors
        input_examples_to_tensors = InputExamplesToTensors(
            self.tokenizer,
            max_seq_length=self.max_seq_length,
            tag_tuple=tuple(tag_list),
            default_logger=self.default_logger,
        )

        _dataloader = dict()
        for phase in input_examples.keys():
            encodings = input_examples_to_tensors(
                input_examples[phase], predict=phase == "predict"
            )

            # dataloader
            data = BertDataset(
                encodings=encodings
            )  # data[j] = 4 torch tensors corresponding to EncodingKeys
            if self.default_logger:
                self.default_logger.log_info(
                    f"[after pre-preprocessing] {phase.ljust(5)} data: {len(data)} examples"
                )

            assert phase in [
                "train",
                "val",
                "test",
                "predict",
            ], f"ERROR! phase = {phase} unknown."
            sampler: Optional[Sampler]
            if phase == "train":
                sampler = RandomSampler(data)
            elif phase in ["val", "test"]:
                sampler = SequentialSampler(data)
            else:  # if phase == "predict":
                sampler = None

            _dataloader[phase] = DataLoader(
                data, sampler=sampler, batch_size=batch_size
            )

        return _dataloader

    ####################################################################################################################
    # HELPER
    ####################################################################################################################
    def _prune_examples(self, list_of_examples: List[Any], phase: str, ratio: float):
        """
        prunes list_of_examples by taking only the first examples
        ---------------------------------------------------------
        :param list_of_examples: [list], e.g. of [InputExample]
        :param phase:            [str], 'train' or 'valid'
        :param (Optional) ratio: [float], e.g. 0.5
        :return: [list], e.g. of [InputExample]
        """
        if ratio == 1.0:
            return list_of_examples
        else:
            num_examples_old = len(list_of_examples)
            num_examples_new = int(ratio * float(num_examples_old))
            info = f"> {phase.ljust(5)} data: use {num_examples_new} of {num_examples_old} examples"
            if self.default_logger:
                self.default_logger.log_info(info)
            return list_of_examples[:num_examples_new]

    @staticmethod
    def _ensure_completeness_in_case_of_bio_tags(tag_list: List[str]) -> List[str]:
        """
        make sure that there is an "I-*" tag for every "B-*" tag in case of BIO-tags
        ----------------------------------------------------------------------------
        :param tag_list:             e.g. ["B-person", "B-time", "I-person"]
        :return: completed_tag_list: e.g. ["B-person", "B-time", "I-person", "I-time"]
        """
        b_tags = [tag for tag in tag_list if tag.startswith("B")]
        for b_tag in b_tags:
            i_tag = b_tag.replace("B-", "I-")
            if i_tag not in tag_list:
                tag_list.append(i_tag)
        return tag_list



from setuptools import setup, find_packages
import pathlib
import os.path
import codecs

BASE_DIR = pathlib.Path(__file__).parent.resolve()


def readme():
    with open(f"{BASE_DIR}/README.rst") as f:
        return f.read()


def requirements():
    with open(f"{BASE_DIR}/requirements.txt") as f:
        return f.read().splitlines()


def requirements_dev():
    with open(f"{BASE_DIR}/requirements_dev.txt") as f:
        return f.read().splitlines()


def get_package_version(rel_path):
    def read(_rel_path):
        here = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(here, _rel_path), "r") as fp:
            return fp.read()

    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="nerblackbox",
    version=get_package_version("nerblackbox/__about__.py"),
    author="Felix Stollenwerk",
    author_email="felix.stollenwerk@arbetsformedlingen.se",
    description="fine-tune transformer-based language models for named entity recognition",
    long_description=readme(),
    long_description_content_type="text/x-rst",
    keywords=[
        "NLP",
        "NER",
        "named entity recognition",
        "BERT",
        "transformer",
        "pytorch",
    ],
    url="https://pypi.org/project/nerblackbox",
    license="Apache 2.0",
    packages=find_packages(),
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    install_requires=requirements(),
    extras_require={
        "dev": requirements_dev(),
    },
    python_requires=">=3.6",
    entry_points="""
            [console_scripts]
            nerbb=nerblackbox.cli:nerbb
        """,
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: Unix",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
    ],
)



import os
from os.path import join

from nerblackbox.modules.datasets.formatter.base_formatter import BaseFormatter


class SUCFormatter(BaseFormatter):
    def __init__(self):
        ner_dataset = "suc"
        ner_tag_list = [
            "person",
            "place",
            "inst",
            "work",
            "misc",
        ]  # misc: see create_ner_tag_mapping()
        super().__init__(ner_dataset, ner_tag_list)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool):
        """
        I: get data
        ----------
        :param verbose: [bool]
        :return: -
        """
        print("(suc: nothing to do)")

    def create_ner_tag_mapping(self):
        """
        II: customize ner_training tag mapping if wanted
        ------------------------------------------------
        :return: ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return dict()
        # return {
        #     'animal': 'misc',
        #     'myth': 'misc',
        #     'product': 'misc',
        #     'event': 'misc',
        #     'other': 'misc',
        # }

    def format_data(self):
        """
        III: format data
        ----------------
        :return: -
        """
        for phase in ["train", "val", "test"]:
            rows = self._read_original_file(phase)
            self._write_formatted_csv(phase, rows)

    def resplit_data(self, val_fraction: float):
        """
        IV: resplit data
        ----------------
        :param val_fraction: [float]
        :return: -
        """
        # train -> train
        df_train = self._read_formatted_csvs(["train"])
        self._write_final_csv("train", df_train)

        # val  -> val
        df_val = self._read_formatted_csvs(["val"])
        self._write_final_csv("val", df_val)

        # test  -> test
        df_test = self._read_formatted_csvs(["test"])
        self._write_final_csv("test", df_test)

    ####################################################################################################################
    # HELPER: READ ORIGINAL
    ####################################################################################################################
    def _read_original_file(self, phase):
        """
        III: format data
        ---------------------------------------------
        :param phase:   [str] 'train' or 'test'
        :return: _rows: [list] of [list] of [str], e.g. [[], ['Inger', 'PER'], ['säger', '0'], ..]
        """
        file_name = {
            "train": "suc-train.conll",
            "val": "suc-dev.conll",
            "test": "suc-test.conll",
        }
        file_path_original = join(self.dataset_path, file_name[phase])

        _rows = list()
        if os.path.isfile(file_path_original):
            with open(file_path_original) as f:
                for i, row in enumerate(f.readlines()):
                    _rows.append(row.strip().split())
            print(f"\n> read {file_path_original}")
        else:
            raise Exception(f"> original file {file_path_original} could not be found.")

        _rows = [
            [row[1], self.transform_tags(row[-3], row[-2])] if len(row) > 0 else list()
            for row in _rows
        ]

        return _rows

    @staticmethod
    def transform_tags(bio, tag):
        """
        :param bio: [str] 'O', 'B', 'I'
        :param tag: [str] '_', 'person', ..
        :return: transformed tag: [str], e.g. 'O', 'B-person', ..
        """
        if bio == "O":
            return "O"
        else:
            return f"{bio}-{tag}"



import warnings
import numpy as np
import torch
from seqeval.metrics import classification_report as classification_report_seqeval
from sklearn.metrics import classification_report as classification_report_sklearn
from typing import List, Dict, Union, Tuple

from nerblackbox.modules.ner_training.metrics.ner_metrics import NerMetrics
from nerblackbox.modules.ner_training.metrics.ner_metrics import convert_to_chunk
from nerblackbox.modules.utils.env_variable import env_variable


class NerModelEvaluation:
    def __init__(
        self,
        current_epoch: int,
        tag_list: List[str],
        dataset_tags: List[str],
        default_logger,
        logged_metrics,
    ):
        self.current_epoch = current_epoch
        self.tag_list = tag_list
        self.dataset_tags = dataset_tags
        self.default_logger = default_logger
        self.logged_metrics = logged_metrics

    ####################################################################################################################
    # 2. VALIDATE / COMPUTE METRICS
    ####################################################################################################################
    def execute(
        self, phase: str, outputs: List[List[torch.Tensor]]
    ) -> Tuple[Dict[str, np.array], str, float]:
        """
        - validate on all batches of one epoch, i.e. whole val or test dataset

        Args:
            phase:   [str], 'val', 'test'
            outputs: [list] of [lists] = [batch_loss, batch_tag_ids, batch_logits] with 3 torch tensors for each batch

        Returns:
            epoch_metrics          [dict] w/ keys 'all_acc', 'fil_f1_micro', .. & values = [np array]
            classification_report: [str]
            epoch_loss:            [float] mean of of all batch losses
        """
        print()
        np_batch: Dict[str, List[np.array]] = self._convert_output_to_np_batch(outputs)
        np_epoch: Dict[
            str, Union[np.number, np.array]
        ] = self._combine_np_batch_to_np_epoch(np_batch)

        # epoch metrics
        epoch_metrics, epoch_tags = self._compute_metrics(phase, np_epoch)

        # classification report
        classification_report = self._get_classification_report(
            phase, self.current_epoch, epoch_tags
        )

        return epoch_metrics, classification_report, np_epoch["loss"]

    @staticmethod
    def _convert_output_to_np_batch(
        outputs: List[List[torch.Tensor]],
    ) -> Dict[str, List[np.array]]:
        """
        - converts pytorch lightning output to np_batch dictionary

        Args:
            outputs: [list] of [lists] = [batch_loss, batch_tag_ids, batch_logits] with 3 torch tensors for each batch

        Returns:
            np_batch: [dict] w/ key-value pairs:
                                'loss':     [list] of <batch_size> x [1D np array]s of length <seq_length>
                                'tag_ids':  [list] of <batch_size> x [1D np array]s of length <seq_length>
                                'logits'    [list] of <batch_size> x [2D np array]s of size   <seq_length x num_tags>
        """
        return {
            "loss": [output[0].detach().cpu().numpy() for output in outputs],
            "tag_ids": [
                output[1].detach().cpu().numpy() for output in outputs
            ],  # [batch_size, seq_length]
            "logits": [
                output[2].detach().cpu().numpy() for output in outputs
            ],  # [batch_size, seq_length, num_tags]
        }

    @staticmethod
    def _combine_np_batch_to_np_epoch(
        np_batch: Dict[str, List[np.array]]
    ) -> Dict[str, Union[np.number, np.array]]:
        """
        - combine np_batch to np_epoch

        Args:
            np_batch: [dict] w/ key-value pairs:
                                'loss':     [list] of <batch_size> x [1D np array]s of length  <seq_length>
                                'tag_ids':  [list] of <batch_size> x [1D np array]s of length  <seq_length>
                                'logits'    [list] of <batch_size> x [2D np array]s of size   [<seq_length>, <num_tags>]

        Returns:
            np_epoch: [dict] w/ key-value pairs:
                                'loss':     [np value]
                                'tag_ids':  [1D np array] of length      <batch_size> x <seq_length>
                                'logits'    [2D np array] of size shape [<batch_size> x <seq_length>, <num_tags>]
        """
        return {
            "loss": np.stack(np_batch["loss"]).mean(),
            "tag_ids": np.concatenate(
                np_batch["tag_ids"]
            ),  # shape: [dataset_size, seq_length]
            "logits": np.concatenate(
                np_batch["logits"]
            ),  # shape: [dataset_size, seq_length, num_tags]
        }

    ####################################################################################################################
    # 1. COMPUTE METRICS ###############################################################################################
    ####################################################################################################################
    def _compute_metrics(
        self, phase: str, _np_epoch: Dict[str, Union[np.number, np.array]]
    ) -> Tuple[Dict[str, np.array], Dict[str, np.array]]:
        """
        - compute loss, acc, f1 scores for size/phase = batch/train or epoch/val-test

        Args:
            phase:         [str], 'train', 'val', 'test'
            _np_epoch: [dict] w/ key-value pairs:
                                 'loss':     [np value]
                                 'tag_ids':  [1D np array] of length      <batch_size> x <seq_length>
                                 'logits'    [2D np array] of size shape [<batch_size> x <seq_length>, <num_tags>]

        Returns:
            _epoch_metrics  [dict] w/ keys 'all_acc', 'fil_f1_micro', .. & values = [np array]
            _epoch_tags     [dict] w/ keys 'true', 'pred'                & values = [np array]
        """
        # batch / dataset
        tag_ids = dict()
        tag_ids["true"], tag_ids["pred"] = self._reduce_and_flatten(
            _np_epoch["tag_ids"], _np_epoch["logits"]
        )

        tags = {
            field: self._convert_tag_ids_to_tags(tag_ids[field])
            for field in ["true", "pred"]
        }
        _epoch_tags = self._get_rid_of_special_tag_occurrences(tags)

        self.default_logger.log_debug("phase:", phase)
        self.default_logger.log_debug(
            "true:", np.shape(tags["true"]), list(set(tags["true"]))
        )
        self.default_logger.log_debug(
            "pred:", np.shape(tags["pred"]), list(set(tags["pred"]))
        )

        if phase == "val":
            for field in ["true", "pred"]:
                np.save(f'{env_variable("DIR_RESULTS")}/{field}.npy', tags[field])

        # batch / dataset metrics
        _epoch_metrics = {"all_loss": _np_epoch["loss"]}
        for tag_subset in [
            "all",
            "fil",
            "chk",
        ] + self.tag_list:
            _epoch_metrics.update(
                self._compute_metrics_for_tags_subset(
                    _epoch_tags, phase, tag_subset=tag_subset
                )
            )

        return _epoch_metrics, _epoch_tags

    @staticmethod
    def _reduce_and_flatten(
        _np_tag_ids: np.array, _np_logits: np.array
    ) -> Tuple[np.array, np.array]:
        """
        helper method for _compute_metrics()
        reduce _np_logits (3D -> 2D), flatten both np arrays (2D -> 1D)

        Args:
            _np_tag_ids: [np array] of shape [batch_size, seq_length]
            _np_logits:  [np array] of shape [batch_size, seq_length, num_tags]

        Returns:
            true_flat: [np array] of shape [batch_size * seq_length], _np_tag_ids               flattened
            pred_flat: [np array] of shape [batch_size * seq_length], _np_logits    reduced and flattened
        """
        true_flat = _np_tag_ids.flatten()
        pred_flat = np.argmax(_np_logits, axis=2).flatten()
        return true_flat, pred_flat

    def _convert_tag_ids_to_tags(self, _tag_ids: np.array) -> np.array:
        """
        helper method for _compute_metrics()
        convert tag_ids (int) to tags (str)

        Args:
            _tag_ids: [np array] of shape [batch_size * seq_length] with [int] elements

        Returns:
            _tags:    [np array] of shape [batch_size * seq_length] with [str] elements
        """
        return np.array(
            [self.tag_list[tag_id] if tag_id >= 0 else "[S]" for tag_id in _tag_ids]
        )

    @staticmethod
    def _get_rid_of_special_tag_occurrences(
        _tags: Dict[str, np.array]
    ) -> Dict[str, np.array]:
        """
        helper method for _compute_metrics()
        get rid of all elements where '[S]' occurs in true array

        Args:
            _tags:      [dict] w/ keys = 'true', 'pred' and
                                  values = [np array] of shape [batch_size * seq_length]

        Returns:
            _tags_new:  [dict] w/ keys = 'true', 'pred' and
                                  values = [np array] of shape [batch_size * seq_length - # of spec. token occurrences]
        """
        pad_indices = np.where(_tags["true"] == "[S]")
        return {key: np.delete(_tags[key], pad_indices) for key in ["true", "pred"]}

    def _compute_metrics_for_tags_subset(
        self, _tags: Dict[str, np.array], _phase: str, tag_subset: str
    ) -> Dict[str, float]:
        """
        helper method for _compute_metrics()
        compute metrics for tags subset (e.g. 'all', 'fil')

        Args:
            _tags:      [dict] w/ keys 'true', 'pred'      & values = [np array]
            _phase:     [str], 'train', 'val'
            tag_subset: [str], e.g. 'all', 'fil', 'PER'

        Returns:
            _metrics    [dict] w/ keys = metric (e.g. 'all_precision_micro') and value = [float]
        """
        tag_list = self._get_filtered_tags(tag_subset)
        tag_group = [tag_subset] if tag_subset in ["all", "fil", "chk"] else ["ind"]
        level = "chunk" if tag_subset == "chk" else "token"

        ner_metrics = NerMetrics(
            _tags["true"],
            _tags["pred"],
            tag_list=tag_list,
            level=level,
            plain_tags=self.dataset_tags == "plain",
        )
        ner_metrics.compute(
            self.logged_metrics.get_metrics(tag_group=tag_group, phase_group=[_phase])
        )
        results = ner_metrics.results_as_dict()

        _metrics = dict()
        # simple
        for metric_type in self.logged_metrics.get_metrics(
            tag_group=tag_group,
            phase_group=[_phase],
            micro_macro_group=["simple"],
            exclude=["loss"],
        ):
            # if results[metric_type] is not None:
            _metrics[f"{tag_subset}_{metric_type}"] = results[metric_type]

        # micro
        for metric_type in self.logged_metrics.get_metrics(
            tag_group=tag_group, phase_group=[_phase], micro_macro_group=["micro"]
        ):
            # if results[f'{metric_type}_micro'] is not None:
            if tag_group == ["ind"]:
                _metrics[f"{tag_subset}_{metric_type}"] = results[
                    f"{metric_type}_micro"
                ]
            else:
                _metrics[f"{tag_subset}_{metric_type}_micro"] = results[
                    f"{metric_type}_micro"
                ]

        # macro
        for metric_type in self.logged_metrics.get_metrics(
            tag_group=tag_group, phase_group=[_phase], micro_macro_group=["macro"]
        ):
            # if results[f'{metric_type}_macro'] is not None:
            _metrics[f"{tag_subset}_{metric_type}_macro"] = results[
                f"{metric_type}_macro"
            ]

        return _metrics

    def _get_filtered_tags(self, _tag_subset: str) -> List[str]:
        """
        helper method for _compute_metrics()
        get list of filtered tags corresponding to _tag_subset name

        Args:
            _tag_subset: [str], e.g. 'all', 'fil', 'PER'

        Returns:
            _filtered_tags: [list] of filtered tags [str]
        """
        if _tag_subset == "all":
            return self.tag_list
        elif _tag_subset in ["fil", "chk"]:
            return [tag for tag in self.tag_list if tag != "O"]
        else:
            return [_tag_subset]

    ####################################################################################################################
    # 2. CLASSIFICATION REPORT #########################################################################################
    ####################################################################################################################
    def _get_classification_report(
        self, phase: str, epoch: int, epoch_tags: Dict[str, np.array]
    ) -> str:
        """
        - get token-based (sklearn) & chunk-based (seqeval) classification report

        Args:
            phase:          [str], 'train', 'val', 'test'
            epoch:          [int]
            epoch_tags:     [dict] w/ keys 'true', 'pred'      & values = [np array]

        Returns:
            classification_report: [str]
        """
        warnings.filterwarnings("ignore")

        # token-based classification report
        tag_list_filtered = self._get_filtered_tags("fil")
        classification_report: str = ""
        classification_report += f"\n>>> Phase: {phase} | Epoch: {epoch}"
        classification_report += (
            "\n--- token-based (sklearn) classification report on fil ---\n"
        )
        classification_report += classification_report_sklearn(
            epoch_tags["true"], epoch_tags["pred"], labels=tag_list_filtered
        )

        # chunk-based classification report
        epoch_tags_chunk = dict()
        for field in ["true", "pred"]:
            epoch_tags_chunk[field] = convert_to_chunk(
                epoch_tags[field], to_bio=self.dataset_tags == "plain"
            )
        self.default_logger.log_debug("> dataset_tags:", self.dataset_tags)
        self.default_logger.log_debug(
            "> epoch_tags_chunk[true]:", list(set(epoch_tags_chunk["true"]))
        )
        self.default_logger.log_debug(
            "> epoch_tags_chunk[pred]:", list(set(epoch_tags_chunk["pred"]))
        )

        classification_report += (
            "\n--- chunk-based (seqeval) classification report on fil ---\n"
        )
        classification_report += classification_report_seqeval(
            epoch_tags_chunk["true"], epoch_tags_chunk["pred"], suffix=False
        )

        warnings.resetwarnings()

        return classification_report



# r"""Python API of the nerblackbox package."""

import os
from os.path import abspath
from typing import Optional, Dict, List, Union
import pandas as pd

try:
    from nerblackbox.modules.main import NerBlackBoxMain
except ModuleNotFoundError as e:
    print(os.getcwd())
    print("- 1 -")
    print(e)
finally:
    try:
        import nerblackbox
    except ModuleNotFoundError as e2:
        print("- 2 -")
        print(e2)


class NerBlackBox:
    """class that provides all nerblackbox functionalities."""

    def __init__(self, base_dir: str = ".", data_dir: str = "./data"):
        """

        Args:
            base_dir: relative path of base directory with respect to current directory
            data_dir: relative path of data directory with respect to current directory

        """

        os.environ["BASE_DIR"] = abspath(base_dir)
        os.environ["DATA_DIR"] = abspath(data_dir)
        print("BASE_DIR = ", os.environ.get("BASE_DIR"))
        print("DATA_DIR = ", os.environ.get("DATA_DIR"))

    ####################################################################################################################
    # NER BLACK BOX
    ####################################################################################################################
    def analyze_data(self, dataset_name: str, **kwargs_optional: Dict):
        """analyze a dataset.

        Args:
            dataset_name: e.g. "swedish_ner_corpus".
            kwargs_optional: with optional key-value pairs {"verbose": [bool]}.
        """

        kwargs = self._process_kwargs_optional(kwargs_optional)
        kwargs["dataset_name"] = dataset_name

        nerbb = NerBlackBoxMain("analyze_data", **kwargs)
        nerbb.main()

    def download(self):
        """download & prepare built-in datasets, prepare experiment configuration.
        needs to be called exactly once before any other CLI/API commands of the package are executed
        in case built-in datasets shall be used.
        """
        _ = self._process_kwargs_optional()

        nerbb = NerBlackBoxMain("download")
        nerbb.main()

    def get_experiment_results(self, experiment_name: str):
        """get results for a single experiment.

        Args:
            experiment_name: e.g. "exp0"
        """
        kwargs = self._process_kwargs_optional()
        kwargs["usage"] = "api"
        kwargs["experiment_name"] = experiment_name

        nerbb = NerBlackBoxMain("get_experiment_results", **kwargs)
        return nerbb.main()

    def get_experiments(self, **kwargs_optional: Dict) -> pd.DataFrame:
        """show list of experiments that have been run.

        Args:
            kwargs_optional: with optional key-value pairs \
            {"ids": [tuple of int], "as_df": [bool]}

        Returns:
            experiments_overview: overview
        """
        kwargs = self._process_kwargs_optional(kwargs_optional)
        kwargs["usage"] = "api"

        nerbb = NerBlackBoxMain("get_experiments", **kwargs)
        return nerbb.main()

    def get_experiments_results(self, **kwargs_optional: Dict) -> Dict:
        """get results from multiple experiments.

        Args:
            kwargs_optional: with optional key-value pairs \
            {"ids": [tuple of int], "as_df": [bool]}

        Returns:
            experiments_results: w/ keys = "best_single_runs", "best_average_runs" \
            & values = [pandas DataFrame] or [dict]
        """
        kwargs = self._process_kwargs_optional(kwargs_optional)
        kwargs["usage"] = "api"
        nerbb = NerBlackBoxMain("get_experiments_results", **kwargs)
        return nerbb.main()

    def init(self):
        """initialize the data_dir directory.
        needs to be called exactly once before any other CLI/API commands of the package are executed.
        """
        _ = self._process_kwargs_optional()

        nerbb = NerBlackBoxMain("init")
        nerbb.main()

    def predict(self, experiment_name: str, text_input: Union[str, List[str]]):
        """predict labels for text_input using the best model of a single experiment.

        Args:
            experiment_name: e.g. "exp0"
            text_input: e.g. "this text needs to be tagged"
        """

        kwargs = self._process_kwargs_optional()
        kwargs["usage"] = "api"
        kwargs["experiment_name"] = experiment_name
        kwargs["text_input"] = text_input

        nerbb = NerBlackBoxMain("predict", **kwargs)
        return nerbb.main()

    def run_experiment(self, experiment_name: str, **kwargs_optional: Dict):
        """run a single experiment.

        Args:
            experiment_name: e.g. "exp0"
            kwargs_optional: with optional key-value pairs \
            {"run_name": [str], "device": [torch device], "fp16": [bool]}
        """

        kwargs = self._process_kwargs_optional(kwargs_optional)
        kwargs["experiment_name"] = experiment_name

        nerbb = NerBlackBoxMain("run_experiment", **kwargs)
        nerbb.main()

    def set_up_dataset(self, dataset_name: str, **kwargs_optional: Dict):
        """set up a dataset using the associated Formatter class.

        Args:
            dataset_name: e.g. "swedish_ner_corpus"
            kwargs_optional: with optional key-value pairs \
            {"modify": [bool], "val_fraction": [float], "verbose": [bool]}
        """

        kwargs = self._process_kwargs_optional(kwargs_optional)
        kwargs["dataset_name"] = dataset_name

        nerbb = NerBlackBoxMain("set_up_dataset", **kwargs)
        nerbb.main()

    def show_experiment_config(self, experiment_name: str):
        """show a single experiment configuration in detail.

        Args:
            experiment_name: e.g. "exp0"
        """
        kwargs = self._process_kwargs_optional()
        kwargs["experiment_name"] = experiment_name

        nerbb = NerBlackBoxMain("show_experiment_config", **kwargs)
        nerbb.main()

    def show_experiment_configs(self):
        """show overview on all available experiment configurations."""
        _ = self._process_kwargs_optional()

        nerbb = NerBlackBoxMain("show_experiment_configs")
        nerbb.main()

    ####################################################################################################################
    # HELPER
    ####################################################################################################################
    @staticmethod
    def _process_kwargs_optional(_kwargs_optional: Optional[Dict] = None):
        if _kwargs_optional is None:
            return {}
        else:
            return {k: v for k, v in _kwargs_optional.items() if v is not None}



r"""This is the nerblackbox package docstring."""
from nerblackbox import __about__
from nerblackbox.api import NerBlackBox
from nerblackbox.modules.experiment_results import ExperimentResults
from nerblackbox.modules.experiments_results import ExperimentsResults
from nerblackbox.modules.ner_training.ner_model_predict import NerModelPredict
